from typing import List
import re
from annotated_dataset.annotation import Annotation


REGEX_EMAIL = re.compile(r"([a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)")


def find_email(text: str, email_label='email') -> List[Annotation]:
    results = []
    for match in re.finditer(REGEX_EMAIL, text):
        results.append(
            Annotation(label=email_label,
                       start=match.span(0)[0],
                       stop=match.span(0)[1]
                       ))
    return results



from pycaprio import Pycaprio


class InceptionClient:
    @staticmethod
    def create_client(host, username, password):
        return Pycaprio(
            inception_host=host,
            authentication=(username, password)
        )



from typing import List, Tuple, Optional
from dataclasses import dataclass
from annotated_dataset.annotation import Annotation


@dataclass
class AnnotatedTextToken:
    start: int
    stop: int


@dataclass
class AnnotatedText:
    text: str
    annotations: List[Annotation]
    resource_id: Optional[str] = None
    resource_index: Optional[int] = None
    tokens: Optional[List[AnnotatedTextToken]] = None

    def get_annotation_signatures(self):
        s = set()
        for annotation in self.annotations:
            s.add(annotation.signature())
        return s

    def get_id(self) -> Tuple[Optional[str], Optional[int]]:
        return self.resource_id, self.resource_index


@dataclass
class AnnotatedTextToken:
    start: int
    stop: int



from typing import List
from annotated_dataset.annotated_text import AnnotatedText
from web_anno_tsv.web_anno_tsv import AnnotatedSentence, Annotation, Span, open_web_anno_tsv


def export_annotated_texts_to_tsv(
        annotated_texts: List[AnnotatedText],
        file: str):

    sentences = []
    for annotated_text in annotated_texts:
        tokens = [Span(
            text=annotated_text.text[token.start: token.stop],
            start=token.start,
            stop=token.stop,
            is_token=True
        ) for token in annotated_text.tokens]
        annotations = [Annotation(
            label=a.label,
            start=a.start,
            stop=a.stop,
            text=annotated_text.text[a.start: a.stop]) for a in annotated_text.annotations]

        sentence = AnnotatedSentence(
            text=annotated_text.text,
            tokens=tokens,
            annotations=annotations
        )
        sentences.append(sentence)

    with open_web_anno_tsv(file, 'w') as f:
        for sentence in sentences:
            f.write(sentence)





from typing import List
from dataclasses import dataclass, field


@dataclass
class NestedAnnotation:
    text: str
    label: str
    start: int
    stop: int
    children: List["NestedAnnotation"] = field(default_factory=list)

    def debug(self, level: int = 0, lines=None):
        if level == 0:
            lines = list()
            lines.append(f'Annotated text: {self.text}')
        indent = '\t'*level
        lines.append(f'{indent}{self.label} [{self.start}:{self.stop}]')
        for child in self.children:
            child.debug(level+1, lines)
        if level == 0:
            return "\n".join(lines)


def iterate(nested_annotation: NestedAnnotation, yield_annotation=False):
    if yield_annotation:
        yield nested_annotation
    for child in nested_annotation.children:
        yield from iterate(child, True)



from typing import Callable
from annotated_dataset.annotation_schema import AnnotationSchema
from annotated_dataset.nested_annotation import NestedAnnotation
from annotated_dataset.annotated_text import AnnotatedText
from annotated_dataset.nested_annotation import iterate
from annotated_dataset.annotated_text_validation_error_report import AnnotatedTextValidationErrorReport


class ValidatorBase:
    def __init__(self, annotation_schema: AnnotationSchema):
        self.annotation_schema = annotation_schema

    def validate_annotated_text(self, annotated_text: AnnotatedText) -> AnnotatedTextValidationErrorReport:
        error_report = AnnotatedTextValidationErrorReport(annotated_text)
        try:
            # Check annotations offsets
            for annotation in annotated_text.annotations:
                with error_report.catch_assertion_error(raise_error=True):
                    assert annotation.start >= 0, 'Annotation start < 0'

                with error_report.catch_assertion_error(raise_error=True):
                    assert annotation.stop <= len(annotated_text.text), 'Annotation stop > len(sentence)'

            # Check labels
            for annotation in annotated_text.annotations:
                with error_report.catch_assertion_error():
                    assert self.annotation_schema.schema.has_node(annotation.label),\
                        f"Invalid label {annotation.label}"

            # Check hierarchy
            nested = self.annotation_schema.nest_annotations(annotated_text)
            self.annotation_schema.valid_children(nested, error_report)

            # Check validator
            for nest_annotation in iterate(nested):
                self.validate_nested_annotation(nest_annotation, error_report)

        except AssertionError:
            pass

        return error_report

    def validate_nested_annotation(self, annotation: NestedAnnotation, error_report: AnnotatedTextValidationErrorReport):
        raise ValueError('Should be implemented')


from networkx import DiGraph
from annotated_dataset.annotation_schema import AnnotationSchema
from annotated_dataset.validator_base import ValidatorBase
from annotated_dataset.nested_annotation import NestedAnnotation
from annotated_dataset.annotated_text_validation_error_report import AnnotatedTextValidationErrorReport
from annotated_dataset.find_email import find_email
from annotated_dataset.find_url import find_url


class ValidatorMapa(ValidatorBase):
    def __init__(self):
        # Create the annotation_schema
        g = DiGraph()

        # Root
        g.add_node('ROOT', category='root')

        # Amount
        g.add_edge('ROOT', 'AMOUNT')
        g.nodes['AMOUNT']['category'] = 'amount'
        for n in [
            'unit',
            'value',
            'other:amount',
            'unresolved:amount'
        ]:
            g.add_edge('AMOUNT', n)
            g.nodes[n]['category'] = 'amount'

        # Date
        g.add_edge('ROOT', 'DATE')
        g.nodes['DATE']['category'] = 'date'
        for n in [
            'day',
            'day of week',
            'month',
            'year',
            'standard abbreviation',
            'calendar event',
            'other:date',
            'unresolved:date'
        ]:
            g.add_edge('DATE', n)
            g.nodes[n]['category'] = 'date'

        g.add_edge('ROOT', 'TIME')
        g.nodes['TIME']['category'] = 'time'

        # Person
        g.add_edge('ROOT', 'PERSON')
        g.nodes['PERSON']['category'] = 'person'
        for n in [
            'title',
            'given name',
            'given name - male',
            'given name - female',
            'family name',
            'family name - male',
            'family name - female',
            'initial name',
            'other:name',
            'unresolved:name',
            'ROLE',
            'PROFESSION'
        ]:
            g.add_edge('PERSON', n)
            g.nodes[n]['category'] = 'person'

        # Address
        g.add_edge('ROOT', 'ADDRESS')
        g.nodes['ADDRESS']['category'] = 'address'
        for n in [
            'street',
            'building',
            'city',
            'place',
            'postcode',
            'territory',
            'country',
            'other:address',
            'unresolved:address'
        ]:
            g.add_edge('ADDRESS', n)
            g.nodes[n]['category'] = 'address'

        # Personal information
        for n in [
            'AGE',
            'ETHNIC CATEGORY',
            'MARITAL STATUS',
            'NATIONALITY',
            'FINANCIAL'
        ]:
            g.add_edge('ROOT', n)
            g.nodes[n]['category'] = 'personal data'

        # Organisation
        g.add_edge('ROOT', 'ORGANISATION')
        g.nodes['ORGANISATION']['category'] = 'organisation'

        # ID
        for n in [
            'id document number',
            'health insurance number',
            'social security number',
            'other:id',
            'unresolved:id',
            'medical record number'
        ]:
            g.add_edge('ROOT', n)
            g.nodes[n]['category'] = 'id'

        # Contact
        for n in [
            'email',
            'telephone number',
            'url',
            'other:contact',
            'unresolved:contact'
        ]:
            g.add_edge('ROOT', n)
            g.nodes[n]['category'] = 'contact'

        # Vehicles
        for n in [
            'build year',
            'colour',
            'licence plate number',
            'model',
            'other:vehicle',
            'type'
        ]:
            g.add_edge('ROOT', n)
            g.nodes[n]['category'] = 'vehicle'

        annotation_schema = AnnotationSchema(g)
        super().__init__(annotation_schema)

    def validate_nested_annotation(self, annotation: NestedAnnotation, error_report: AnnotatedTextValidationErrorReport) -> None:
        # Label
        if annotation.label in {'AMOUNT', 'DATE', 'PERSON', 'ADDRESS'}:
            with error_report.catch_assertion_error():
                assert len(annotation.children) >= 1, \
                    f"{_a(annotation)} is expected to have at least one child"

        if annotation.label == 'AMOUNT':
            has_value = False
            for sub_annotation in annotation.children:
                if sub_annotation.label in {'value',  'other:amount', 'unresolved:amount'}:
                    has_value = True
                    break
            with error_report.catch_assertion_error():
                assert has_value, f'{_a(annotation)} is expected to have a sub-annotation ' \
                                  f'`value` or `other:amount` or `unresolved:amount`'

        elif annotation.label == 'PERSON':
            counts = self._count_sub(annotation)
            num_family_name = counts['family name'] + counts['family name - male'] + counts['family name - female']
            with error_report.catch_assertion_error():
                assert num_family_name in {0, 1}, f"{num_family_name} `family name` has been found"

            num_given_name = counts['given name'] + counts['given name - male'] + counts['given name - female']
            with error_report.catch_assertion_error():
                assert num_given_name in {0, 1}, f"{num_given_name} `given name` has been found"

            with error_report.catch_assertion_error():
                assert num_family_name + num_given_name + counts['initial name'] > 0,\
                    f'{_a(annotation)} should have at least a `family name` , ' \
                    f'a `given name` or an `initials name` tag'

        elif annotation.label == 'email':
            emails = find_email(annotation.text)
            found = False
            for email in emails:
                if email.start == annotation.start and email.stop == annotation.stop:
                    found = True
                    break
            assert found, f'Incorrect `{_a(annotation)}`'

        elif annotation.label == 'url':
            urls = find_url(annotation.text)
            found = False
            for url in urls:
                if url.start == annotation.start and url.stop == annotation.stop:
                    found = True
                    break
            assert found, f'Incorrect `{_a(annotation)}`'

    def _count_sub(self, annotation: NestedAnnotation):
        d = dict()
        for child in self.annotation_schema.get_children(annotation.label):
            d[child] = 0
        for sub_annotation in annotation.children:
            if sub_annotation.label in d:
                d[sub_annotation.label] += 1
        return d


def _a(annotation: NestedAnnotation):
    return f'`{annotation.label}[{annotation.start}:{annotation.stop}]`'



from typing import NamedTuple, List, Tuple, Optional
from pathlib import Path
from itertools import groupby
import logging
from annotated_dataset.annotated_resource_base import AnnotatedResourceBase
from annotated_dataset.annotated_text import AnnotatedText
from annotated_dataset.validator_base import ValidatorBase
from annotated_dataset.annotation_schema import AnnotationSchema
from collections import defaultdict
from bokeh.io import output_file, save
from bokeh.plotting import figure
from bokeh.transform import factor_cmap
from bokeh.palettes import d3
from bokeh.layouts import column
from bokeh.models import FactorRange, ColumnDataSource, DataTable, TableColumn, Paragraph, Div
from annotated_dataset.annotated_text_validation_error_report import AnnotationError
from annotated_dataset.export_annotated_texts_to_xmi import export_annotated_texts_to_xmi
from annotated_dataset.export_annotated_texts_to_tsv import export_annotated_texts_to_tsv


class GoldCorpusGroup:
    def __init__(self, annotated_texts: List[AnnotatedText]):
        self.annotated_texts = annotated_texts

    def get_annotated_texts(self):
        return self.annotated_texts

    def get_annotated_text_ids(self) -> List[Tuple[Optional[str], Optional[int]]]:
        return list(map(lambda a: a.get_id(), self.annotated_texts))


class GoldCorpus:
    def __init__(self, groups: List[GoldCorpusGroup]):
        self.groups = groups
        self.index = dict()
        for group in self.groups:
            for annotated_text_id in group.get_annotated_text_ids():
                self.index[annotated_text_id] = group

    @staticmethod
    def create_from_dataset(dataset: "AnnotatedDataset"):
        groups = []
        annotated_texts = list()
        for resource in dataset.resources:
            annotated_texts.extend(resource.get_annotated_texts())
        sorted_by_text = sorted(annotated_texts, key=lambda annotated_text: annotated_text.text)

        for _, group in groupby(sorted_by_text, key=lambda annotated_text: annotated_text.text):
            group = list(group)
            if len(group) > 1:
                groups.append(GoldCorpusGroup(group))
        return GoldCorpus(groups)

    def get_groups(self) -> List[GoldCorpusGroup]:
        return self.groups

    def is_gold(self, annotated_text: AnnotatedText) -> bool:
        return annotated_text.get_id() in self.index

    def get_group(self, annotated_text: AnnotatedText):
        annotated_text_id = annotated_text.get_id()
        if annotated_text_id in self.index:
            return self.index[annotated_text_id]
        return None


class GoldCorpusPreferredResources:
    def __init__(self, preferred_resources: List[str]):
        self.preferred_resources = preferred_resources

    def get_preferred_in_group(self, group: GoldCorpusGroup):
        for annotated_text in group.get_annotated_texts():
            for preferred_resource in self.preferred_resources:
                if annotated_text.resource_id == preferred_resource:
                    return annotated_text

        for annotated_text in group.get_annotated_texts():
            for preferred_resource in self.preferred_resources:
                if annotated_text.resource_id.find(preferred_resource):
                    return annotated_text

        return group.get_annotated_texts()[0]


class ValidationError(NamedTuple):
    annotated_text: AnnotatedText
    error: AnnotationError


class AnnotatedDataset:
    def __init__(self, name: str, annotated_resources: List[AnnotatedResourceBase]):
        self.name = name
        self.resources = annotated_resources
        self.gold_corpus: GoldCorpus = GoldCorpus.create_from_dataset(self)

    def validate(self, validator: ValidatorBase):
        errors = []
        for resource in self.resources:
            annotated_texts = resource.get_annotated_texts()
            for i, annotated_text in enumerate(annotated_texts):
                report = validator.validate_annotated_text(annotated_text)
                for error in report.errors:
                    errors.append(ValidationError(
                        annotated_text=annotated_text,
                        error=error
                    ))
        return errors

    def get_stats(self, annotation_schema: AnnotationSchema):
        stats = defaultdict(dict)
        for category, labels in annotation_schema.get_labels_by_category().items():
            for label in labels:
                stats[category][label] = 0
        for resource in self.resources:
            annotated_texts = resource.get_annotated_texts()
            for annotated_text in annotated_texts:
                for annotation in annotated_text.annotations:
                    if annotation_schema.is_valid_label(annotation.label):
                        category = annotation_schema.get_label_category(annotation.label)
                        stats[category][annotation.label] += 1
        return stats

    def create_report(self, validator: ValidatorBase, file="report.txt"):
        widgets = list()

        # Main title
        widgets.append(Div(text=f'<h1>Dataset analysis for {self.name}</h1>'))

        # Resource widget
        widgets.extend(self._get_resource_widget())

        # Annotation label distribution widget
        widgets.extend(self._get_annotation_label_distribution_widget(validator))

        # Annotation errors widget
        widgets.extend(self._get_annotation_error_widget(validator))

        # Gold
        widgets.extend(self._get_gold_widget())

        # Save
        output_file(file)
        save(column(*widgets), file)

    def _get_resource_widget(self):
        resources = []
        num_annotated_texts = []
        num_annotations = []

        for resource in self.resources:
            resources.append(resource.resource_id)
            annotated_texts = resource.get_annotated_texts()
            num_annotated_texts.append(len(annotated_texts))
            num_annotations.append(sum([len(at.annotations) for at in annotated_texts]))

        data = dict(
            resource=resources,
            sentences=num_annotated_texts,
            annotations=num_annotations
        )
        source = ColumnDataSource(data)

        columns = [
            TableColumn(field="resource", title="Resource", width=500),
            TableColumn(field="sentences", title="Sentences", width=70),
            TableColumn(field="annotations", title="Annotations", width=880),
        ]

        data_table = DataTable(
            source=source,
            columns=columns,
            min_height=100,
            max_height=1000,
            height=(len(resources) + 1) * 28,
            width=1500,
            autosize_mode="none",
            sizing_mode="scale_both")

        return [Div(text='<h2>Resources</h2>'), data_table ]

    def _get_annotation_label_distribution_widget(self, validator):
        stats = self.get_stats(validator.annotation_schema)

        labels = []
        occurrences = []
        categories = []
        max_value = 0
        category_and_label = []
        for category, stats_by_label in stats.items():
            if category != 'root':
                categories.append(category)
                for label, num in stats_by_label.items():
                    label = f"{label} ({num})"
                    labels.append(label)
                    occurrences.append(num)
                    if num > max_value:
                        max_value = num
                    category_and_label.append((category, label))

        palette = d3['Category20'][len(categories)]

        source = ColumnDataSource(data=dict(category_and_label=category_and_label, occurrences=occurrences))

        tag_set_graph = figure(
            x_range=FactorRange(*category_and_label),
            plot_width=1500,
            plot_height=600,
            title="Dataset",
            toolbar_location=None,
            tools="")

        tag_set_graph.vbar(
            x='category_and_label',
            top='occurrences',
            width=0.9,
            source=source,
            line_width=0,
            line_color="white",
            fill_color=factor_cmap('category_and_label', palette=palette, factors=categories, start=0, end=1))

        tag_set_graph.y_range.start = 0
        tag_set_graph.x_range.range_padding = 0.1
        tag_set_graph.xaxis.major_label_orientation = 'vertical'
        tag_set_graph.xaxis.major_label_text_font_size = "12px"
        tag_set_graph.xaxis.group_label_orientation = 'vertical'
        tag_set_graph.xaxis.group_text_font_size = "12px"

        return [Div(text=f'<h2>Tag distribution</h2>'), tag_set_graph]

    def _get_annotation_error_widget(self, validator):
        widgets = [Div(text=f'<h2>Annotation errors</h2>')]

        # Validate dataset
        for resource, errors in groupby(self.validate(validator), key=lambda e: e.annotated_text.resource_id):
            widgets.append(Paragraph(text=f"Error in {resource}"))
            line_numbers = []
            texts = []
            error_types = []
            error_messages = []
            for error in errors:
                line_numbers.append(error.annotated_text.resource_index)
                texts.append(error.annotated_text.text)
                error_types.append(str(error.error.type).replace('ErrorType.', ''))
                error_messages.append(error.error.message)

            data = dict(
                line_numbers=line_numbers,
                texts=texts,
                error_types=error_types,
                error_messages=error_messages
            )
            source = ColumnDataSource(data)

            columns = [
                TableColumn(field="line_numbers", title="Line nº", width=50),
                TableColumn(field="error_types", title="Error type", width=70),
                TableColumn(field="error_messages", title="Error", width=400),
                TableColumn(field="texts", title="Text", width=940)
            ]

            data_table = DataTable(
                source=source,
                columns=columns,
                min_height=100,
                max_height=1000,
                height=(len(line_numbers) + 1) * 28,
                width=1500,
                autosize_mode="none",
                sizing_mode="scale_both")

            widgets.append(data_table)

        return widgets

    def _get_gold_widget(self):
        widgets = [Div(text='<h2>Gold annotations</h2>')]

        # Compare annotations
        column_groups = []
        column_resource_ids = []
        column_resource_indices = []
        column_diffs = []
        column_texts = []

        compare_id = 0
        for group in self.gold_corpus.get_groups():
            annotation_signatures_in_annotated_texts = []
            for annotated_text in group.get_annotated_texts():
                annotation_signatures_in_annotated_texts.append(
                    annotated_text.get_annotation_signatures())
            common_annotations_signatures = set.intersection(*annotation_signatures_in_annotated_texts)

            has_diff = False
            diff_in_annotated_texts = []
            for annotation_signatures_in_annotated_text in annotation_signatures_in_annotated_texts:
                diffs = []
                for signature in annotation_signatures_in_annotated_text:
                    if signature not in common_annotations_signatures:
                        has_diff = True
                        diffs.append(signature)
                diff_in_annotated_texts.append(diffs)

            if has_diff:
                for annotated_text, diffs in zip(
                        group.get_annotated_texts(),
                        diff_in_annotated_texts):
                    column_groups.append(f'#{compare_id}')
                    column_resource_ids.append(annotated_text.resource_id)
                    column_resource_indices.append(annotated_text.resource_index)
                    column_texts.append(annotated_text.text)
                    column_diffs.append(", ".join([f'{d[0]}[{d[1]}:{d[2]}]' for d in diffs]))
                compare_id += 1

        data = dict(
            group=column_groups,
            resource_ids=column_resource_ids,
            resource_indices=column_resource_indices,
            texts=column_texts,
            diffs=column_diffs
        )
        source = ColumnDataSource(data)

        columns = [
            TableColumn(field="group", title="Group", width=50),
            TableColumn(field="resource_ids", title="Resource", width=250),
            TableColumn(field="resource_indices", title="Line nº", width=50),
            TableColumn(field="diffs", title="Diff", width=400),
            TableColumn(field="texts", title="Text", width=400)
        ]

        data_table = DataTable(
            source=source,
            columns=columns,
            min_height=100,
            max_height=1000,
            height=(len(column_groups) + 1) * 28,
            width=1500,
            autosize_mode="none",
            sizing_mode="scale_both")

        widgets.append(data_table)

        return widgets

    def get_gold_and_not_gold(self, gold_corpus_preferred_resources: GoldCorpusPreferredResources):
        gold = list()
        not_gold = list()

        for resource in self.resources:
            for annotated_text in resource.get_annotated_texts():
                if self.gold_corpus.is_gold(annotated_text):
                    gold_group = self.gold_corpus.get_group(annotated_text)
                    preferred = gold_corpus_preferred_resources.get_preferred_in_group(gold_group)
                    if preferred == annotated_text:
                        gold.append(annotated_text)
                else:
                    not_gold.append(annotated_text)
        return gold, not_gold

    def create_bundle(self,
                      bundle_dir: str,
                      validator: ValidatorBase,
                      gold_corpus_preferred_resources: GoldCorpusPreferredResources
                      ):

        if not self.resources:
            raise ValueError('No resource found')

        bundle_dir = Path(bundle_dir.replace(' ', '_'))
        bundle_dir.mkdir()

        # Create report
        logging.info('Create report')
        report_file = bundle_dir.joinpath(f'{bundle_dir.name}_report.html')
        self.create_report(validator, str(report_file))
        logging.info(' ---> done')

        logging.info('Export data')
        type_system = self.resources[0].type_system
        gold, not_gold = self.get_gold_and_not_gold(gold_corpus_preferred_resources)

        # Export .xmi
        data_xmi_file = bundle_dir.joinpath(f'{bundle_dir.name}_data.zip')
        export_annotated_texts_to_xmi(
            not_gold, type_system, data_xmi_file, xmi_file=f'{bundle_dir.name}_data.xmi')

        # Export .tsv
        data_tsv_file = bundle_dir.joinpath(f'{bundle_dir.name}_data.tsv')
        export_annotated_texts_to_tsv(not_gold, str(data_tsv_file))

        logging.info(' ---> done')

        logging.info('Export gold data')
        # Export  gold .xmi
        gold_xmi_file = bundle_dir.joinpath(f'{bundle_dir.name}_data_gold.zip')
        export_annotated_texts_to_xmi(
            gold, type_system, gold_xmi_file, xmi_file=f'{bundle_dir.name}_data_gold.xmi')

        # Export .tsv
        gold_tsv_file = bundle_dir.joinpath(f'{bundle_dir.name}_data_gold.tsv')
        export_annotated_texts_to_tsv(gold, str(gold_tsv_file))
        logging.info(' ---> done')

        logging.info(f'Exportation finished `{bundle_dir}`', )


from typing import List
from annotated_dataset.annotated_text import AnnotatedText, AnnotatedTextToken
from annotated_dataset.annotation import Annotation
from annotated_dataset.annotated_resource_base import AnnotatedResourceBase
from dkpro_cassis_tools import load_cas_from_zip_file, \
    restore_cas_segmentation_by_newline, \
    SENTENCE_NS, \
    NAMED_ENTITY_NS, \
    TOKEN_NS


class AnnotatedResourceUimaCasXmi(AnnotatedResourceBase):
    def __init__(self, path, restore_segmentation_by_newline=False, resource_id=None):
        self.path = path
        self.restore_segmentation_by_newline = restore_segmentation_by_newline
        if not resource_id:
            resource_id = path

        super().__init__(resource_id)

        # Open Resource
        with open(self.path, 'rb') as f:
            self.cas, self.type_system = load_cas_from_zip_file(f, return_type_system=True)

        # Extract annotated texts
        self.annotated_texts = []
        if self.restore_segmentation_by_newline:
            self.cas = restore_cas_segmentation_by_newline(self.cas)
        for i, sentence in enumerate(self.cas.select(SENTENCE_NS)):
            # Text
            text = sentence.get_covered_text()

            # Annotations
            annotations = []
            for a in self.cas.select_covered(NAMED_ENTITY_NS, sentence):
                annotations.append(Annotation(
                    label=a.value,
                    start=a.begin - sentence.begin,
                    stop=a.end - sentence.begin

                ))

            # Tokens
            tokens = []
            for token in self.cas.select_covered(TOKEN_NS, sentence):
                tokens.append(AnnotatedTextToken(
                    start=token.begin-sentence.begin,
                    stop=token.end-sentence.begin
                ))
            not_spaces = set()
            for token in tokens:
                for j in range(token.start, token.stop):
                    not_spaces.add(j)
            for j, char in enumerate(text):
                if j not in not_spaces:
                    if not text[j].isspace():
                        # TODO
                        print("Tokenizing Error")

            self.annotated_texts.append(AnnotatedText(
                text=text,
                annotations=annotations,
                resource_id=self.resource_id,
                resource_index=i + 1,
                tokens=tokens
            ))

    def get_annotated_texts(self) -> List[AnnotatedText]:
        return self.annotated_texts




from typing import List
from annotated_dataset.annotated_text import AnnotatedText


class AnnotatedResourceBase:
    def __init__(self, resource_id: str):
        self.resource_id = resource_id

    def get_annotated_texts(self) -> List[AnnotatedText]:
        raise ValueError("get_annotated_texts should be implemented")



from typing import List
from annotated_dataset.annotated_text import AnnotatedText
from cassis import Cas
from dkpro_cassis_tools import TOKEN_NS, SENTENCE_NS, NAMED_ENTITY_NS
from dkpro_cassis_tools import dump_cas_to_zip_file


def export_annotated_texts_to_xmi(
        annotated_texts: List[AnnotatedText],
        type_system,
        file: str,
        xmi_file=None):

    cas = Cas(typesystem=type_system)

    current_start = 0
    starts = []
    sofa_string = ''

    # Create sofa string
    for annotated_text in annotated_texts:
        starts.append(current_start)
        text = annotated_text.text
        if not text.endswith('\n'):
            text += '\n'
        sofa_string += text
        current_start += len(text)

    cas.sofa_string = sofa_string

    # Tokens
    for annotated_text, start in zip(annotated_texts, starts):
        for token in annotated_text.tokens:
            annotation = cas.typesystem.get_type(TOKEN_NS)(
                begin=start + token.start,
                end=start + token.stop)
            cas.add_annotation(annotation)

    # Sentences
    for annotated_text, start in zip(annotated_texts, starts):
        annotation = cas.typesystem.get_type(SENTENCE_NS)(
            begin=start,
            end=start + len(annotated_text.text))
        cas.add_annotation(annotation)

    # Annotations
    for annotated_text, start in zip(annotated_texts, starts):
        for annotation in annotated_text.annotations:
            annotation = cas.typesystem.get_type(NAMED_ENTITY_NS)(
                value=annotation.label,
                begin=start + annotation.start,
                end=start + annotation.stop)
            cas.add_annotation(annotation)

    # write
    with open(file, 'wb') as f:
        dump_cas_to_zip_file(cas, f, xmi_file=xmi_file)




from typing import List, Tuple, Dict
from networkx import DiGraph, shortest_path_length, NetworkXNoPath
from annotated_dataset.annotation import Annotation
from annotated_dataset.annotated_text import AnnotatedText
from annotated_dataset.nested_annotation import NestedAnnotation
from annotated_dataset.annotated_text_validation_error_report import AnnotatedTextValidationErrorReport
from _collections import defaultdict


class AnnotationSchema:
    def __init__(self, schema: DiGraph, root_node="ROOT"):
        self.schema = schema
        self.root_node = root_node

        if root_node not in schema:
            raise ValueError(f'Root node `{root_node}` node is missing')

        self.labels: List[str]
        self.shortest_path_from_root: Dict[str, int] = dict()
        for node in schema.nodes:
            if node != root_node:
                try:
                    self.shortest_path_from_root[node] = shortest_path_length(schema, root_node, node)
                except NetworkXNoPath:
                    raise

    def valid_children(self, nested_annotation, error_report: AnnotatedTextValidationErrorReport):
        for child in nested_annotation.children:
            with error_report.catch_assertion_error():
                assert self.schema.has_edge(nested_annotation.label, child.label), \
                    f"Invalid hierarchy {nested_annotation.label}->{child.label}"
            self.valid_children(child, error_report)

    def nest_annotations(self, annotated_text: AnnotatedText) -> NestedAnnotation:
        nested_annotations_index = []
        nested_annotations = []
        sorted_annotations = sorted(annotated_text.annotations, key=self.order_annotations_fn)

        for annotation in sorted_annotations:
            nested_annotation = NestedAnnotation(
                text=annotated_text.text,
                label=annotation.label,
                start=annotation.start,
                stop=annotation.stop
            )

            ancestors = [a for a in nested_annotations_index
                         if a.start <= annotation.start
                         and a.stop >= annotation.stop]
            if not ancestors:
                nested_annotations_index.append(nested_annotation)
                nested_annotations.append(nested_annotation)
            else:
                nested_annotations_index.append(nested_annotation)
                ancestors[-1].children.append(nested_annotation)

        return NestedAnnotation(
            text=annotated_text.text,
            label=self.root_node,
            start=0,
            stop=len(annotated_text.text),
            children=nested_annotations)

    def order_annotations_fn(self, annotation: Annotation) -> Tuple[int, int, int]:
        d = self.shortest_path_from_root[annotation.label] \
            if annotation.label in self.shortest_path_from_root else 1000
        return (
            annotation.start,
            -annotation.stop,
            d)

    def get_children(self, node: str) -> List[str]:
        nodes = []
        for n in self.schema.neighbors(node):
            nodes.append(n)
        return nodes

    def get_labels_by_category(self):
        categories = defaultdict(list)
        for n in self.schema.nodes:
            category = self.schema.nodes[n]['category']
            categories[category].append(n)
        return categories

    def get_label_category(self, label):
        return self.schema.nodes[label]['category']

    def is_valid_label(self, label):
        return label in self.schema.nodes




import tempfile
from annotated_dataset.annotated_dataset import AnnotatedDataset
from annotated_dataset.validator_mapa import ValidatorMapa
from annotated_dataset.inception_project import InceptionProject
from annotated_dataset.annotated_dataset import GoldCorpusPreferredResources
from pathlib import Path
import logging


def export(config):
    # Extract resource
    resources = []
    with tempfile.TemporaryDirectory() as tmp:
        tmp = Path(tmp)
        for project_config in config['inception_projects']:
            # Create the project
            project = InceptionProject.create_from_remote(
                project_config['name'],
                project_config['inception_client'],
                str(tmp)
            )

            # Extract resources
            resources.extend(project.extract_resources(
                restore_segmentation_by_newline=project_config['use_segmentation_by_newline']))

    # Create dataset
    dataset_name = config['dataset_name']
    dataset = AnnotatedDataset(
        dataset_name,
        resources)

    # Create validator
    validator = ValidatorMapa()

    # Create gold_corpus_preferred_resources
    gold_corpus_preferred_resources = GoldCorpusPreferredResources(
        config['gold_corpus_preferred_resources']
    )

    # Create bundle
    bundle_dir = Path(config['export_dir']).joinpath('export_' + dataset_name.lower())
    dataset.create_bundle(
        str(bundle_dir),
        validator,
        gold_corpus_preferred_resources
    )







from typing import NamedTuple


class Annotation(NamedTuple):
    label: str
    start: int
    stop: int

    def __hash__(self):
        return self.signature()

    def signature(self):
        return self.label, self.start, self.stop




from typing import List
from contextlib import contextmanager
from annotated_dataset.annotated_text import AnnotatedText
import enum


class ErrorType(enum.Enum):
    Error = 1
    Warning = 2


class AnnotationError:
    def __init__(self, error_type: ErrorType, message):
        self.type = error_type
        self.message = message

    def __repr__(self):
        return f'{str(self.type)}: {self.message}'


class AnnotatedTextValidationErrorReport:
    def __init__(self, annotated_text: AnnotatedText):
        self.annotated_text = annotated_text
        self.errors: List[AnnotationError] = []

    @contextmanager
    def catch_assertion_error(self, error_type: ErrorType = ErrorType.Error, raise_error=False):
        try:
            yield
        except AssertionError as e:
            self.errors.append(AnnotationError(error_type, str(e)))
            if raise_error:
                raise



from typing import List
from pathlib import Path
import zipfile
from pycaprio import Pycaprio
from pycaprio.mappings import InceptionFormat
from annotated_dataset.annotated_resource_uima_cas_xmi import AnnotatedResourceUimaCasXmi
from annotated_dataset.annotated_resource_base import AnnotatedResourceBase
import logging


class ProjectNotFound(Exception):
    def __init__(self, project_name: str, inception_client: Pycaprio):
        super().__init__(f'Project `{project_name}` in {inception_client} ')


class InceptionProject:
    def __init__(self, path):
        self.path = Path(path)
        if not self.path.is_dir():
            raise ValueError(f'Dir `{path}` not found')

    def extract_resources(self, restore_segmentation_by_newline=False) -> List[AnnotatedResourceBase]:
        resources = []
        for path in sorted(self.path.glob('annotation/*/*.zip')):
            resource_id = self.path.name + '/' + path.parent.name
            resources.append(AnnotatedResourceUimaCasXmi(
                str(path),
                resource_id=resource_id,
                restore_segmentation_by_newline=restore_segmentation_by_newline))
        return resources

    @staticmethod
    def create_from_remote(project_name: str, inception_client: Pycaprio, location: str):
        # Get project id
        project_id = None
        projects = inception_client.api.projects()
        for project in projects:
            if project.project_name == project_name:
                project_id = project.project_id
                break

        if project_id is None:
            raise ProjectNotFound(project_name, inception_client)

        # Download project zip
        logging.info(f'Download project `{project_name}`')
        zip_file = Path(location).joinpath(f"{project_name}.zip")
        with open(zip_file, 'wb') as z:
            zip_content = inception_client.api.export_project(
                project_id,
                project_format=InceptionFormat.XMI)
            z.write(zip_content)

        # Unzip project
        project_dir = Path(location).joinpath(project_name)
        with zipfile.ZipFile(zip_file, 'r') as z:
            z.extractall(project_dir)

        # Create the project
        project = InceptionProject(str(Path(location).joinpath(project_name)))
        logging.info(f' ---> Done')

        return project




import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="annotated_dataset",
    version="0.0.7",
    author="Laurent Bié",
    author_email="l.bie@pangeanic.com  ",
    description="Annotated dataset analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pangeamt/annotated_dataset",
    packages=setuptools.find_packages(),
    install_requires=[
        'dkpro-cassis>=0.5.0',
        'dkpro-cassis-tools>=0.0.3',
        'networkx>=2.5',
        'pycaprio>=0.2.0',
        'web_anno_tsv>=0.0.1',
        'bokeh>=2.2.3'
    ],
    tests_require=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],

    python_requires='>=3.7',
)




from annotated_dataset.annotation_schema import AnnotationSchema
from annotated_dataset.validator_base import ValidatorBase
from networkx import DiGraph
from annotated_dataset.nested_annotation import NestedAnnotation
from annotated_dataset.annotated_text_validation_error_report import AnnotatedTextValidationErrorReport


class ValidatorAda(ValidatorBase):
    def __init__(self):
        # Create the annotation_schema
        g = DiGraph()
        g.add_node('ROOT')
        g.add_edge('ROOT', 'PERSON')
        g.add_edge('ROOT', 'LOCATION')
        g.add_edge('ROOT', 'DATE')
        g.add_edge('ROOT', 'AMOUNT')
        g.add_edge('ROOT', 'EMAIL')
        g.add_edge('ROOT', 'PHONE-NUMBER')
        g.add_edge('ROOT', 'URL')
        g.add_edge('ROOT', 'ORGANISATION')
        g.add_edge('ROOT', 'TIME')
        g.add_edge('ROOT', 'ID')
        g.add_edge('ROOT', 'ORG')
        g.add_edge('PERSON', 'person-title')
        g.add_edge('PERSON', 'person-given-name')
        g.add_edge('PERSON', 'person-given-name-male')
        g.add_edge('PERSON', 'person-given-name-female')
        g.add_edge('PERSON', 'person-family-name')
        g.add_edge('PERSON', 'person-family-name-male')
        g.add_edge('PERSON', 'person-family-name-female')
        g.add_edge('LOCATION', 'location-place')
        g.add_edge('LOCATION', 'location-street')
        g.add_edge('LOCATION', 'location-postcode')
        g.add_edge('LOCATION', 'location-city')
        g.add_edge('LOCATION', 'location-territory')
        g.add_edge('DATE', 'date-day')
        g.add_edge('DATE', 'date-day-of-week')
        g.add_edge('DATE', 'date-month')
        g.add_edge('DATE', 'date-year')
        g.add_edge('DATE', 'date-standard-abbreviation')
        g.add_edge('DATE', 'date-calendar-event')
        g.add_edge('AMOUNT', 'amount-unit')
        g.add_edge('AMOUNT', 'amount-value')
        g.add_edge('location-place', 'PERSON')
        g.add_edge('ORG', 'PERSON')

        annotation_schema = AnnotationSchema(g)
        super().__init__(annotation_schema)

    def validate_nested_annotation(self, annotation: NestedAnnotation, error_report: AnnotatedTextValidationErrorReport) -> None:
        # Label
        if annotation.label in {'PERSON', 'LOCATION', 'DATE', 'AMOUNT'}:
            with error_report.catch_assertion_error():
                assert len(annotation.children) >= 1, \
                    f"`{annotation.label}` is expected to have at least one child"
            if annotation.label == 'AMOUNT':
                has_value = False
                for sub_annotation in annotation.children:
                    if sub_annotation.label == 'amount-value':
                        has_value = True
                        break
                with error_report.catch_assertion_error():
                    assert has_value, f'AMOUNT is expected to have a sub-annotation amount-value'









